/*
  frag.c

  Dug Song <dugsong@anzen.com>

  Copyright (c) 1999 Anzen Computing. All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. All advertising materials mentioning features or use of this software
     must display the following acknowledgement:
     This product includes software developed by Anzen Computing.
  4. Neither the name of Anzen Computing nor the names of its
     contributors may be used to endorse or promote products derived
     from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
  ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  $Id: frag.c,v 1.14 1999/05/04 15:14:37 dugsong Exp $
*/

#include "config.h"

#ifdef STDC_HEADERS
#include <stdlib.h>
#endif
#include <libnet.h>
#include "frag.h"
#include "send.h"

#ifndef MIN
#define MIN(x,y)        ((x) < (y) ? (x) : (y))
#endif

int frag_preserve_headers = 0;

int
frag_init(int preserve_headers)
{
  frag_preserve_headers = preserve_headers;
  return 1;
}

/*
  frag_insert
  Args: pointer to fragment, body and len of frag to follow.
  Return: pointer to appended fragment.
*/
FRAG *
frag_insert(FRAG *frag, u_char *data, int fraglen)
{
  FRAG *new;

  if ((new = malloc(sizeof(FRAG))) == NULL)
    return NULL;

  if (!frag) {
    new->data = data;
    new->len = fraglen;
    new->head = new;
    new->prev = NULL;
    new->next = NULL;
  }
  else {
    new->data = data;
    new->len = fraglen;
    new->head = frag->head;
    new->next = frag->next;
    new->prev = frag;
    frag->next = new;
    if (new->next) new->next->prev = new;
  }
  return (new);
}

/*
  frag_make
  Args: pointer to packet, packet length, fragment size.
  Return: pointer to start of packet fragment list.
*/
FRAG *
frag_make(u_char *pkt, int pktlen, int fragsize)
{
  /* Yuck foo. */
  FRAG *frag = NULL;
  struct ip *iph = (struct ip *)pkt;
  int ip_hl = iph->ip_hl * 4;
  u_char ip_tos = iph->ip_tos;
  u_short ip_len = ntohs(iph->ip_len);
  u_short ip_id = ip_id = ntohs(iph->ip_id);
  u_char ip_ttl = iph->ip_ttl;
  u_char ip_p = iph->ip_p;
  u_long ip_src = iph->ip_src.s_addr;
  u_long ip_dst = iph->ip_dst.s_addr;
  u_char *ip_blob = pkt + ip_hl;
  u_char *ip_end = pkt + ip_len;
  u_char *p, *data;
  short ip_fragoff;
  int len;

  /* Make sure fragmentation is valid for this packet. */
  if (pktlen <= ip_hl + fragsize || fragsize % 8 != 0)
    return NULL;
  
  /* Fragment packet. */
  for (p = ip_blob ; p < ip_end ; p += len) {
    if (p == ip_blob && frag_preserve_headers) {
      if (ip_end - p <= 24) return NULL;
      len = 24;
    }
    else
#ifdef BSD
    /* XXX - weird OpenBSD sendto() bug: < 4 byte payload returns ENOBUFS? */
    if ((ip_end - p) <= (fragsize + 4))
      len = ip_end - p;
    else
#endif /* BSD */
    len = MIN(ip_end - p, fragsize);

    ip_fragoff = ((p + len < ip_end) ? IP_MF : 0) | ((p - ip_blob) >> 3);

    if (libnet_init_packet(IP_H + len, &data) == -1)
      return NULL;
    
    libnet_build_ip(0, ip_tos, ip_id, ip_fragoff, ip_ttl, ip_p,
		    ip_src, ip_dst, p, len, data);
    
    if (!(frag = frag_insert(frag, data, IP_H + len)))
      return NULL;
  }
  return (frag->head);
}

FRAG *
frag_add_fwd_overwrite(FRAG *frag)
{
  FRAG *f;
  u_char nulls[IP_MAXPACKET], *newdata;
  int newoff, newlen;

  memset(nulls, 0, sizeof(nulls));
  
  for (f = frag ; f && f->next ; f = f->next) {
    struct ip *iph = (struct ip *)f->data;
    int ip_hl = iph->ip_hl * 4;
    u_char ip_tos = iph->ip_tos;
    u_short ip_len = ntohs(iph->ip_len);
    u_short ip_id = ip_id = ntohs(iph->ip_id);
    u_char ip_ttl = iph->ip_ttl;
    u_char ip_p = iph->ip_p;
    u_long ip_src = iph->ip_src.s_addr;
    u_long ip_dst = iph->ip_dst.s_addr;

    /* Build new null data frag, half the length of the current frag,
       and covering the latter half of the current frag. */
    newlen = (ip_len - ip_hl) / 2;
    newoff = ntohs(iph->ip_off) + (newlen >> 3);

    if (libnet_init_packet(IP_H + newlen, &newdata) == -1)
      return NULL;
    
    libnet_build_ip(0, ip_tos, ip_id, newoff, ip_ttl, ip_p,
		    ip_src, ip_dst, nulls, newlen, newdata);

    /* Add null fragment after current fragment. */
    if (!frag_insert(f, newdata, IP_H + newlen))
      return NULL;

    /* Swap them. */
    if (!frag_swap(f))
      return NULL;
  }
  return frag;
}
    
int
frag_send(FRAG *frag)
{
  FRAG *f;

  for (f = frag ; f != NULL ; f = f->next) {
    if (!send_packet(f->data, f->len))
      return 0;
  }
  return 1;
}

/* Destroy this, and all following fragments. */
int
frag_free(FRAG *frag)
{
  FRAG *next, *f = frag;

  if (f->prev) f->prev->next = NULL;
  
  for ( ; f != NULL ; f = next) {
    next = f->next;
    libnet_destroy_packet(&(f->data));
    free(f);
  }
  return 1;
}

FRAG *
frag_last(FRAG *frag)
{
  FRAG *f;

  if (!frag) return NULL;
  for (f = frag ; f->next != NULL ; f = f->next) ;
  return (f);
}
  
FRAG *
frag_dup(FRAG *frag)
{
  FRAG *dup = frag_insert(frag, frag->data, frag->len);
  return (dup);
}

/*
  frag_swap
  Args: pointer to fragment a (followed by b).
  Return: pointer to same fragment a (after swap with b).
*/
FRAG *
frag_swap(FRAG *frag)
{
  FRAG *f, *next = frag->next;

  if (next) {
    frag->next = next->next;
    next->prev = frag->prev;
    frag->prev = next;
    next->next = frag;

    if (frag->next) frag->next->prev = frag;

    if (next->prev) next->prev->next = next;
    else {
      for (f = next ; f != NULL ; f = f->next)
	f->head = next;
    }
  }
  else if (frag->prev == frag->head) { /* last frag of two. */
    return frag_swap(frag->head);
  }
  else { /* last frag of a few, swap with head. */
    next = frag->head;
    next->prev = frag->prev;
    frag->prev = NULL;
    frag->next = next->next;
    next->next = NULL;
    if (frag->next)
      frag->next->prev = frag;
    if (next->prev)
      next->prev->next = next;
    
    for (f = frag ; f != NULL ; f = f->next)
      f->head = frag;
  }
  return (frag);
}
